#include <stdlib.h>
#include "syspro.h"
#include "sysprotransform.h"
#include "sysprolinear.h"

/* test for the syspro linear package
   one preprocessor used: scaling
   this uses PETSc's KSP for actually solving a linear system
*/

/* this example features a preprocessor that we can loop over:
   preconditioners
*/
#undef __FUNCT__
#define __FUNCT__ "create_solver"
/*! Create a solver and 
  install a monitor that dynamically increases the maximum
  number of iterations.
*/
static PetscErrorCode create_solver(NumericalProblem prob,void **ctx)
{
  MPI_Comm comm; KSP solver; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = NumericalProblemGetComm(prob,&comm); CHKERRQ(ierr);
  ierr = KSPCreate(comm,&solver); CHKERRQ(ierr);
  *(KSP*)ctx = solver;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "destroy_solver"
static PetscErrorCode destroy_solver(void *ctx)
{
  KSP solver; PetscErrorCode ierr;
  PetscFunctionBegin;
  solver = (KSP)ctx;
  ierr = KSPDestroy(&solver); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "setup_pc_choices"
static PetscErrorCode setup_pc_choices()
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = NewTransformObject("pc",PCNONE,PETSC_NULL); CHKERRQ(ierr);
  ierr = NewTransformObject("pc",PCJACOBI,PETSC_NULL); CHKERRQ(ierr);
  ierr = NewTransformObject("pc",PCILU,PETSC_NULL); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "setup_pc"
static PetscErrorCode setup_pc
(const char *type,int pcv,PetscBool overwrite,
 NumericalProblem inproblem,NumericalProblem *outproblem,
 void *gctx,void **ctx,PetscBool *success)
{
  LinearSystem problem = (LinearSystem)inproblem;
  KSP solver; PC pc;  Mat A;
  PetscErrorCode ierr;

  PetscFunctionBegin;
  *outproblem = inproblem;

  /*
   * the fun part is setting up the Petsc PC
   */
  ierr = LinearSystemGetParts
    (problem,&A,PETSC_NULL,PETSC_NULL,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
  solver = (KSP)gctx;
  ierr = KSPSetOperators(solver,A,A,SAME_NONZERO_PATTERN); CHKERRQ(ierr);
  ierr = KSPGetPC(solver,&pc); CHKERRQ(ierr);
  ierr = PCSetType(pc,type); CHKERRQ(ierr);
  *success = PETSC_TRUE;

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "unset_pc"
static PetscErrorCode unset_pc
(const char *type,PetscBool overwrite,void *ctx,void *gctx,
 NumericalProblem thisproblem,NumericalProblem upproblem,
 NumericalSolution old,NumericalSolution nnew)
{
#if 0
  LinearSystem problem = (LinearSystem)thisproblem;
  Mat A;
  PetscErrorCode ierr;
#endif
  PetscFunctionBegin;

#if 0
  ierr = LinearSolutionCopy
    ((LinearSolution)old,(LinearSolution)nnew); CHKERRQ(ierr);
#endif

  PetscFunctionReturn(0);
}


#undef __FUNCT__
#define __FUNCT__ "solvelinear"
static PetscErrorCode solvelinear
(NumericalProblem problem,void *dum,NumericalSolution *rsol)
{
  LinearSystem linear = (LinearSystem)problem;
  KSP solver; Mat A; Vec rhs,sol; 
  KSPConvergedReason reason; PetscInt it; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = LinearSystemGetParts(linear,&A,NULL,&rhs,NULL,NULL); CHKERRQ(ierr);
  ierr = VecDuplicate(rhs,&sol); CHKERRQ(ierr);
  ierr = KSPCreate(MPI_COMM_SELF,&solver); CHKERRQ(ierr);
  ierr = KSPSetOperators(solver,A,A,DIFFERENT_NONZERO_PATTERN); CHKERRQ(ierr);
  ierr = KSPSolve(solver,rhs,sol); CHKERRQ(ierr);
  ierr = KSPGetConvergedReason(solver,&reason); CHKERRQ(ierr);
  if (reason<=0) SETERRQ(PETSC_COMM_SELF,1,"This should have converged");
  ierr = KSPGetIterationNumber(solver,&it); CHKERRQ(ierr);
  printf("Convergence in %d iterations\n",it);
  ierr = KSPDestroy(&solver); CHKERRQ(ierr);
  ierr = LinearCreateNumericalSolution(NULL,rsol); CHKERRQ(ierr);
  ierr = LinearSolutionSetVector((LinearSolution)*rsol,sol); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

int main(int argc,char **argv) {
  NumericalProblem problem; NumericalSolution solution;
  Mat A; Vec rhs,sol; PetscErrorCode ierr;
  PetscInitialize(&argc,&argv,0,0);
  ierr = SysProInitialize(); CHKERRQ(ierr);

  /* set up a simple problem */
  {
    LinearSystem linearproblem;
    int n=100,i; PetscScalar two=2.,mone=-1.,one=1.;

#include "testmat.c"
    /*
     * now create the linear system to contain them
     */
    ierr = CreateLinearSystem(MPI_COMM_SELF,&linearproblem); CHKERRQ(ierr);
    ierr = LinearSystemSetParts
      (linearproblem,A,NULL,rhs,sol,NULL); CHKERRQ(ierr);
    problem = (NumericalProblem)linearproblem;
  }

  /* declare the scalings */
  ierr = DeclareScalingPreprocessor(); CHKERRQ(ierr);
  ierr = PetscOptionsSetValue("-syspro_scaling","exhaustive"); CHKERRQ(ierr);

  /* declare a preconditioner loop */
  ierr = DeclarePreprocessor
    (/* name */ "pc",
     /* one-time creation */ &setup_pc_choices,
     /* creation and deletion for specific choices */ NULL,NULL,NULL,
     /* create/delete context that will live for the duration of this
	preprocessor */
     &create_solver,&destroy_solver,
     /* set and unset general/specific ? choice */
     &setup_pc,NULL /*&unset_pc*/ ); CHKERRQ(ierr);
  /* and make sure you loop over it */
  ierr = PetscOptionsSetValue("-syspro_pc","none"); CHKERRQ(ierr);

  /* declare functions */
  ierr = SysProDeclareFunctions
    (NULL,NULL,NULL, /* general setup functions */
     solvelinear,LinearDeleteNumericalProblem, /* problem functions */
     &LinearCreateNumericalSolution,NULL,LinearDeleteNumericalSolution, /* solution functions */
     NULL,NULL,NULL /* context functions */
     ); CHKERRQ(ierr);

  ierr = PreprocessorsOptionsHandling(); CHKERRQ(ierr);

  /* give some feedback on what's happening */
  ierr = SysProDeclareTraceFunction(SysProDefaultTrace); CHKERRQ(ierr);

  /* solve this problem */
  ierr = PreprocessedProblemSolving(problem,&solution); CHKERRQ(ierr);

  /* check for correctness */
  {
    Vec sol,asol; int mone=-1; PetscReal nrm;
    ierr = LinearSolutionGetVector
      ((LinearSolution)solution,&sol); CHKERRQ(ierr);
    ierr = VecDuplicate(sol,&asol); CHKERRQ(ierr);
    ierr = MatMult(A,sol,asol); CHKERRQ(ierr);
    ierr = VecAXPY(asol,mone,rhs); CHKERRQ(ierr);
    ierr = VecNorm(asol,NORM_2,&nrm); CHKERRQ(ierr);
    if (nrm>1.e-6)
      SETERRQ1(PETSC_COMM_SELF,1,"Computed wrong solution: norm residual is %e",nrm);
    ierr = VecDestroy(&asol); CHKERRQ(ierr);
  }
  ierr = LinearDeleteNumericalSolution(solution); CHKERRQ(ierr);
  ierr = MatDestroy(&A); CHKERRQ(ierr);
  ierr = VecDestroy(&rhs); CHKERRQ(ierr);
  ierr = VecDestroy(&sol); CHKERRQ(ierr);
  ierr = PetscFree(problem); CHKERRQ(ierr);
  ierr = SysProFinalize(); CHKERRQ(ierr);
  PetscFinalize();
  PetscFunctionReturn(0);
}
